/*
 * Wrapper around libsocket++'s ftp class for inclusion in Tcl
 *
 * There's a bug in the ftp code of libsocket. Use the patch at the
 * end of this file against libsocket++ 1.10
 *
 * The ftp class writes all its output to an ostream. We intercept
 * this data in our ostrstream, and pass back the contents of this
 * output as return value to Tcl.
 */

#include <tcl.h>
#include <tclobj.h>
#include <iostream.h>
#include <strstream.h>
#include <ftp.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>

/*
 * Wrapper class
 */

class tclftp : public ftp, public Tcl_Object {
public:
  ~tclftp ();

  tclftp       (Tcl_Interp *, int, char *[], int);
  int Tcl_Proc (Tcl_Interp *, int, char *[], int);
  TCL_OBJECT_DECL_CLASS(tclftp);

private:
  int connected;
  ostrstream output;
};

/*
 * commands
 */

static const TclObj_Comtab fftab[] = {
  /* constructors */
  {  1, "new",   "",       "create unconnected ftp object" },

  /* operators */
  {  1, "open",  "string", "connect to host and log in as anonymous" },
  {  2, "open",  "string string string",
     "connect to host and log in as user with password" },
  {  3, "close", "",       "close connection" },
  {  4, "cd",    "string", "change working directory on remote server" },
  {  5, "pwd",   "",       "show current remote directory" },
  {  6, "dir",   "",       "get directory listing" },
  {  7, "get",   "string", "get remote file" },
  {  9, "put",   "string", "upload file" },
  {  8, "get",   "string string", "get remote file as" },
  { 10, "put",   "string string", "upload file as" },
  {  0, NULL, NULL, NULL }
};

/*
 * Declare ourselves
 */

TCL_OBJECT_DECL_BODY(tclftp,Tcl_Object,fftab,NULL);

/*
 * Tcl-Style constructor
 */

tclftp::tclftp (Tcl_Interp *interp, int argc, char *argv[], int command)
  : ftp (&output)
{
  connected = 0;
}

tclftp::~tclftp (void)
{
  if (connected)
    rdbuf()->quit();
}

int
tclftp::Tcl_Proc (Tcl_Interp *interp,
		  int argc, char *argv[],
		  int command)
{
  ftp::replycodea code;
  char *s1, *s2, *s3;

  output.seekp(0);

  switch (command) {
  case 1: /* open string */
    {
      struct passwd *pwent = getpwuid (getuid());
      char address[1024];

      sprintf (address, "%s@", (pwent)?pwent->pw_name:"nobody");

      if (rdbuf()->connect(TCL_OBJECT_ARG_STRING(interp, argv[2])) != 0) {
	code = ftp::rca_error;
	break;
      }

      connected = 1;

      if ((code = rdbuf()->get_response())          >= ftp::rca_error) break;
      if ((code = rdbuf()->user    ("anonymous"))   >= ftp::rca_error) break;
      if ((code = rdbuf()->passwd  (address))       >= ftp::rca_error) break;
      if ((code = rdbuf()->rep_type(ftp::rt_image)) >= ftp::rca_error) break;
    }
    break;
  case 2: /* open string string string */
    TclObj_CastArgs (interp, argc-2, argv+2,
		     "string string string",
		     &s1, &s2, &s3);

    if (rdbuf()->connect(s1) != 0) {
      code = ftp::rca_error;
      break;
    }

    connected = 1;

    if ((code=rdbuf()->get_response())          >= ftp::rca_error) break;
    if ((code=rdbuf()->user    (s2))            >= ftp::rca_error) break;
    if ((code=rdbuf()->passwd  (s3))            >= ftp::rca_error) break;
    if ((code=rdbuf()->rep_type(ftp::rt_image)) >= ftp::rca_error) break;
    break;

  case 3: /* close */
    code = rdbuf()->quit();
    connected = 0;
    break;

  case 4: /* cd */
    code = rdbuf()->cd (TCL_OBJECT_ARG_STRING(interp, argv[2]));
    break;

  case 5: /* pwd */
    code = rdbuf()->pwd ();
    break;

  case 6: /* dir */
    code = rdbuf()->list ();
    break;

  case 7: /* get string */
    code = rdbuf()->getfile (TCL_OBJECT_ARG_STRING(interp, argv[2]), NULL);
    break;

  case 8: /* get string string */
    code = rdbuf()->getfile (TCL_OBJECT_ARG_STRING(interp, argv[2]),
			     TCL_OBJECT_ARG_STRING(interp, argv[3]));
    break;

  case 9: /* put string */
    code = rdbuf()->putfile (TCL_OBJECT_ARG_STRING(interp, argv[2]), NULL);
    break;

  case 10: /* put string string */
    code = rdbuf()->putfile (TCL_OBJECT_ARG_STRING(interp, argv[2]),
			     TCL_OBJECT_ARG_STRING(interp, argv[3]));
    break;
  }
  /*
   * return what has accumulated in output. flush last character.
   */

  if (output.tellp) output.seekp(-1,cur);
  output << (char) 0;

  Tcl_SetResult (interp,
		 output.str(),
		 TCL_VOLATILE);
  output.freeze(0);

  if (code>=ftp::rca_error)
    return TCL_ERROR;

  return TCL_OK;
}

/*
 * Patch to libsocket++ 1.10 to avoid sending a callback address of
 * 127.0.0.1 to the server <g>. While at it, don't depend on getting
 * local port 10000 but accept any. Also fix a compiler warning.
 *

diff -r -c socket++-1.10.orig/ftp.C socket++-1.10/ftp.C
*** socket++-1.10.orig/ftp.C	Wed Oct 18 01:07:34 1995
--- socket++-1.10/ftp.C	Wed Oct 23 22:10:32 1996
***************
*** 60,66 ****
  {
    sockinetbuf sb (sockbuf::sock_stream, 0);
  
!   if (sb.bind_until_success (portno)) {
      sock_error ("ftpdata: ", "bind_until_success failed");
      exit (1);
    }
--- 60,66 ----
  {
    sockinetbuf sb (sockbuf::sock_stream, 0);
  
!   if (sb.bind_until_success ((portno,0))) {
      sock_error ("ftpdata: ", "bind_until_success failed");
      exit (1);
    }
***************
*** 175,181 ****
  
  ftp::replycodea ftp::ftpbuf::useraddr (sockinetaddr sa)
  {
!   if (sa.sin_addr.s_addr == 0) {
      // local host
      char hostname [64];
      if (::gethostname (hostname, 63) == -1) {
--- 175,181 ----
  
  ftp::replycodea ftp::ftpbuf::useraddr (sockinetaddr sa)
  {
!   if (sa.sin_addr.s_addr,0 == 0) {
      // local host
      char hostname [64];
      if (::gethostname (hostname, 63) == -1) {
diff -r -c socket++-1.10.orig/sockstream.h socket++-1.10/sockstream.h
*** socket++-1.10.orig/sockstream.h	Wed Oct 18 01:07:29 1995
--- socket++-1.10/sockstream.h	Wed Oct 23 20:29:07 1996
***************
*** 61,67 ****
--- 61,69 ----
  #	define SO_RCVTIMEO	0x1006
  
  #	define MSG_MAXIOVLEN	16
+ #ifndef SOMAXCONN
  #	define SOMAXCONN	5
+ #endif
  #endif // __linux__
  
  extern void sock_error (const char* classname, const char* error_message);

 *
 * End of patch
 */
